SystemOrganization addCategory: #'O-Languages-LRSTM'!
SystemOrganization addCategory: #'O-Languages-AmbientTalk-Reflection'!
SystemOrganization addCategory: #'O-Languages-ActiveObject'!
SystemOrganization addCategory: #'O-Languages-ActiveObject-Reflection'!
SystemOrganization addCategory: #'O-Languages-Clojure'!
SystemOrganization addCategory: #'O-Languages-Clojure-Reflection'!
SystemOrganization addCategory: #'O-Languages-CSPi-Reflection'!
SystemOrganization addCategory: #'O-Languages-CSPi'!
SystemOrganization addCategory: #'O-Languages-Debugging'!

OstDomain subclass: #ATActor
	instanceVariableNames: 'process mailbox terminate name'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'O-Languages-AmbientTalk-Reflection'!

!ATActor commentStamp: 'StefanMarr 4/22/2012 20:01' prior: 0!
The main idea behind the implementation of ATActors is that the Smalltalk semantics already provides encapsulation on object level.
Thus, we can rely on redefining the requestExection handlers, but do not need to implement anything else.
This provides us the whole actor semantics since the requestExection handler will turn calls into async requests.


TODO: literals/globals are still a problematic loop hole. We do not consider them at the moment. oops...
We probably should make a local copy of first request and then use that exclusively, or something like that. But here we are really fighting against our Smalltalk heritage, and well, we could make it safe...!

!ATActor class methodsFor: 'constructor' stamp: 'StefanMarr 3/22/2012 16:59'!
newWithInitializer: aBlock
	"Normally, we would return a reference to the domain,
	 but this is not really the way AmbientTalk works,
	 so, here we also return the value returned by the aBlock."
	| domain valueMsg promise |
	domain := self new.
	promise := domain evaluateAsync: [aBlock omni: domain].
	^ promise read.! !

!ATActor methodsFor: 'actor semantics' stamp: 'StefanMarr 3/31/2012 22:25'!
evaluateAsync: aBlock
	"This is an implementation level method.
	 The block is by convention not referring to
	 unsafe shared state."
	| msg |
	msg := ATMessage new.
	msg receiver: aBlock.
	msg setSelector: #value arguments: #().

	mailbox nextPut: msg.
	
	^ msg promise.! !

!ATActor methodsFor: 'initialization' stamp: 'StefanMarr 1/4/2012 13:58'!
initialize
	super initialize.
	
	mailbox := SharedQueue new.
	terminate := false.
	
	process := self spawnHere: [
		[terminate] whileFalse: [
			self processIncomingMessages.
		]].
	process name: 'Actor'.! !

!ATActor methodsFor: 'debugging' stamp: 'StefanMarr 1/4/2012 13:43'!
name
	^ name! !

!ATActor methodsFor: 'debugging' stamp: 'StefanMarr 1/4/2012 13:58'!
name: aString
	name := aString.
	
	process ifNotNil: [ process name: 'Actor [', aString, ']' ].! !

!ATActor methodsFor: 'printing' stamp: 'StefanMarr 1/4/2012 13:54'!
printOn: aStream
	super printOn: aStream.
	
	name ifNotNil: [
		aStream nextPutAll: '[', name, ']'.].! !

!ATActor methodsFor: 'actor semantics' stamp: 'StefanMarr 1/23/2013 13:35'!
processIncomingMessages
	| msg result |
	<omniUnenforced>
	msg := mailbox next.
	
	[ result := msg send.
	  msg result: result ]
		on: Error
		do: [:err | msg failed: err ]! !

!ATActor methodsFor: 'actor semantics' stamp: 'StefanMarr 1/3/2012 14:44'!
sendAsAsyncMessage: aSelector to: anObj with: anArgArray
	| msg |
	msg := ATMessage new.
	msg setSelector: aSelector arguments: anArgArray.
	msg receiver: anObj.
	mailbox nextPut: msg.
	
	^ msg result.! !

!ATActor methodsFor: 'actor semantics' stamp: 'StefanMarr 1/23/2013 12:54'!
terminate
	<omniUnenforced>
	^ self evaluateAsync: [ terminate := true ]! !

OstDomain subclass: #CSPiProcess
	instanceVariableNames: 'process state completionSema'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'O-Languages-CSPi'!

!CSPiProcess methodsFor: 'initialization' stamp: 'StefanMarr 7/28/2012 10:59'!
initialize
	super initialize.
	completionSema := Semaphore new.
	state := #initialized.! !

!CSPiProcess methodsFor: 'process' stamp: 'StefanMarr 1/9/2012 23:46'!
join
	"Wait for completion"
	completionSema wait.! !

!CSPiProcess methodsFor: 'process' stamp: 'StefanMarr 1/9/2012 23:54'!
raiseIsolationViolationError
	Error signal: 'Entity is isolated. Access violation detected.'! !

!CSPiProcess methodsFor: 'process runnable' stamp: 'StefanMarr 1/9/2012 23:18'!
run
	self subclassResponsibility! !

!CSPiProcess methodsFor: 'process' stamp: 'StefanMarr 7/18/2012 22:08'!
start
	<omniUnenforced>
	process := self spawnHere: [
		state := #running.
		[self run] omni: self.
		state := #done.
		completionSema signal.
	].
	process name: 'CSPi'.! !

OstDomain subclass: #HaltDomain
	instanceVariableNames: ''
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'O-Languages-Debugging'!

!HaltDomain class methodsFor: 'defaults' stamp: 'StefanMarr 8/3/2012 22:59'!
default
	Default ifNil: [Default := self new].
	^ Default! !

OstDomain subclass: #LRSTMDomain
	instanceVariableNames: ''
	classVariableNames: 'Domain'
	poolDictionaries: ''
	category: 'O-Languages-LRSTM'!

!LRSTMDomain class methodsFor: 'singleton' stamp: 'StefanMarr 12/11/2011 22:19'!
domain
	"We only need one domain for the STM system."
	Domain ifNil: [ Domain := self new. ].
	^ Domain! !

!LRSTMDomain class methodsFor: 'initialization' stamp: 'StefanMarr 7/31/2012 22:43'!
initialize
	Domain := nil.! !

OstDomain subclass: #OActiveObjectDomain
	instanceVariableNames: 'obj'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'O-Languages-ActiveObject-Reflection'!

!OActiveObjectDomain class methodsFor: 'instance creation' stamp: 'StefanMarr 1/9/2012 17:17'!
newFor: anActiveObject
	| domain |
	domain := self new.
	domain object: anActiveObject.
	^ domain.! !

!OActiveObjectDomain methodsFor: 'omni-domain' stamp: 'StefanMarr 11/6/2011 21:04'!
domainForNewObjects
	"Active objects are created explicitly,
	 classes tell whether their objects are supposed to be active.
	 This happens during #initialize.
	 
	 However, #domainForNewObjects is determined during
	 #basicNew, and for ActiveObjects we just assume as the standard
	 case that all objects are passive.
	 For the moment that means, they are not part of any domain."
	^ nil! !

!OActiveObjectDomain methodsFor: 'accessing' stamp: 'StefanMarr 11/5/2011 20:27'!
object: anObject
	obj := anObject! !

OstDomain subclass: #OcjAgentDomain
	instanceVariableNames: 'agent'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'O-Languages-Clojure-Reflection'!

!OcjAgentDomain class methodsFor: 'instance creation' stamp: 'StefanMarr 1/6/2012 13:09'!
newFor: anAgent
	| domain |
	domain := self new.
	domain agent: anAgent.
	^ domain.! !

!OcjAgentDomain methodsFor: 'accessing' stamp: 'StefanMarr 1/5/2012 14:42'!
agent: anAgent
	agent := anAgent! !

!OcjAgentDomain methodsFor: 'agent semantics' stamp: 'StefanMarr 7/1/2013 12:39'!
handleExecReq: aSelector args: argArray on: anObject
	<omniUnenforced>
	| normSel |
	
	"Rules are only enforced on the agent itself"
	anObject = agent ifFalse: [
		^ anObject baselevelPerform: aSelector withArguments: argArray].
	
	normSel := aSelector asNormalSelector.
	
	(normSel = #read or:   [
	 normSel = #await or:  [
	 normSel = #shutdown   ]])
		ifTrue: [
		^ agent perform: normSel	].
	
	(normSel = #send:) ifTrue: [
		^ agent send: (argArray at: 1)].
	
	(normSel = #send:with:) ifTrue: [
		^ agent send: (argArray at: 1) with: (argArray at: 2)].
	
	"There are two more selectors, we do not want to have someone messing with:"
	(normSel = #processIncomingMessages or: [
	 normSel = #initialize])
		ifTrue: [
			Error signal: 'Access to implementation methods is denied.'].
	
	"For everything else, we just assume someone added an explicit protocal to a Smalltalk subclass.
	 Thus, we follow Agent semantics by doing it asynchronously, but we do not give a return value.
	 Could be done with a promise, but we will encorage here asynchnous/agent-like programming."
	^ agent perform: #send: with: [:oldState | agent baselevelPerform: aSelector withArguments: argArray.
														oldState].! !

OstDomain subclass: #OcjImmutableDomain
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'O-Languages-Clojure-Reflection'!

!OcjImmutableDomain methodsFor: 'immutability' stamp: 'StefanMarr 1/6/2012 15:14'!
raiseImmutabilityError
	Error signal: 'Attempt to modify the object failed. The owning domain does not allow mutations.'! !

Message subclass: #ATMessage
	instanceVariableNames: 'promise receiverObj'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'O-Languages-AmbientTalk-Reflection'!

!ATMessage class methodsFor: 'instance creation' stamp: 'StefanMarr 11/6/2011 22:52'!
method: aCompiledMethod on: anObj with: arguments
	| msg |
	msg := self new.
	msg method: aCompiledMethod.
	msg receiver: anObj.
	msg arguments: arguments.
	^ msg! !

!ATMessage methodsFor: 'message' stamp: 'StefanMarr 1/23/2013 13:33'!
failed: anError
	promise failed: anError! !

!ATMessage methodsFor: 'initialize-release' stamp: 'StefanMarr 1/4/2012 14:11'!
initialize
	super initialize.
	promise := OPromise new.! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/6/2011 22:53'!
promise
	^ promise! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/6/2011 22:52'!
receiver
	^ receiverObj! !

!ATMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/6/2011 22:52'!
receiver: anObj
	receiverObj := anObj! !

!ATMessage methodsFor: 'requirements' stamp: 'StefanMarr 1/2/2012 16:52'!
result
	"Result is blocking on the promise.
	 Use #promise to obtain the promise directly."
	^ promise read! !

!ATMessage methodsFor: 'message' stamp: 'StefanMarr 11/6/2011 22:52'!
result: anObj
	promise deliver: anObj! !

!ATMessage methodsFor: 'message' stamp: 'StefanMarr 3/23/2012 13:30'!
send
	"answer the result of sending this message to its receiver"

	^ receiverObj baselevelPerform: selector withArguments: args! !

Message subclass: #OActiveObjectMessage
	instanceVariableNames: 'promise receiverObj'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'O-Languages-ActiveObject-Reflection'!

!OActiveObjectMessage class methodsFor: 'instance creation' stamp: 'StefanMarr 11/5/2011 17:17'!
method: aCompiledMethod on: anObj with: arguments
	| msg |
	msg := self new.
	msg method: aCompiledMethod.
	msg receiver: anObj.
	msg arguments: arguments.
	^ msg! !

!OActiveObjectMessage methodsFor: 'initialization' stamp: 'StefanMarr 11/4/2011 17:23'!
initialize
	super initialize.
	promise := OPromise new.! !

!OActiveObjectMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/5/2011 23:17'!
receiver
	^ receiverObj! !

!OActiveObjectMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/5/2011 17:19'!
receiver: anObj
	receiverObj := anObj! !

!OActiveObjectMessage methodsFor: 'promise' stamp: 'StefanMarr 11/4/2011 17:24'!
resolve
	^ promise read.! !

!OActiveObjectMessage methodsFor: 'accessing' stamp: 'StefanMarr 1/9/2012 17:48'!
result
	^ promise! !

!OActiveObjectMessage methodsFor: 'accessing' stamp: 'StefanMarr 11/5/2011 23:19'!
result: anObj
	promise deliver: anObj! !

!OActiveObjectMessage methodsFor: 'active object' stamp: 'StefanMarr 7/1/2013 12:34'!
send
	"answer the result of sending this message to its receiver"
	^ receiverObj baselevelPerform: selector withArguments: args! !

!LRSTMTransaction methodsFor: '*O-Languages-LRSTM' stamp: 'StefanMarr 8/2/2012 14:49'!
do: aBlock ifConflict: aConflictBlock
	"Evaluate aBLock within a new transaction, unless we are already in an existing transactional context. Evaluate aConflictBlock if the transaction conflicts with concurrent edits."

	<omniUnenforced>
	| result |
	parent := Processor activeProcess currentTransaction.
	Processor activeProcess currentTransaction: self.
	context := thisContext sender.
	self begin.
	
	"We need to use #evaluateInContext: here,
	 instead of #omni: to avoid undesired transformation"
	[ [result := LRSTMDomain domain evaluateInContextOnBaseLevel: aBlock ]
		ensure: [ Processor activeProcess currentTransaction: parent ]
	] ifCurtailed: [ self abort ].
 
	self commitIfConflict: aConflictBlock.
	^ result! !

OstObject subclass: #OActiveObject
	instanceVariableNames: 'mailbox terminate process'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'O-Languages-ActiveObject'!

!OActiveObject commentStamp: 'StefanMarr 1/9/2012 16:12' prior: 0!
An active object enforces that all operations on it
are executed under its own control, i.e., in its own thread(Process).

All non-active objects are call passive objects and do not impose any
restrictions. Thus, messages are processes in the context of the sending
active/passive object. Passive objects do not have their own thread of control.

All non-active objects are passive objects, i.e., only objects inheriting from
ActiveObject are to be considered active objects.
!

!OActiveObject methodsFor: 'initialization' stamp: 'StefanMarr 7/18/2012 21:57'!
initialize
	<omniUnenforced>
	| domain |
	super initialize.
	
	mailbox := SharedQueue new.
	terminate := false.
	
	domain := OActiveObjectDomain newFor: self.
	OstMirror setDomainOf: self to: domain.
	
	process := domain newProcessFor: [
		[terminate and: [mailbox isEmpty]] whileFalse: [
			self processIncomingMessages.
		]].
	process name: 'OActiveObject'.
	process resume.! !

!OActiveObject methodsFor: 'active object' stamp: 'StefanMarr 7/18/2012 22:03'!
processIncomingMessages
	<omniUnenforced>
	| msg result |
	msg := mailbox next.
	
	result := msg send.
	
	msg result: result.! !

!OActiveObject methodsFor: 'active object' stamp: 'StefanMarr 1/9/2012 18:36'!
sendAsAsyncMessage: aSelector to: anObj with: anArgArray
	| msg |
	msg := OActiveObjectMessage new.
	msg setSelector: aSelector arguments: anArgArray.
	msg receiver: anObj.
	
	mailbox nextPut: msg.
	
	^ msg result.! !

OstObject subclass: #OcjAgent
	instanceVariableNames: 'state mailbox terminate process'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'O-Languages-Clojure'!

!OcjAgent class methodsFor: 'instance creation' stamp: 'StefanMarr 7/18/2012 22:01'!
new: initialState
	"Work around omni enforcement"
	<omniUnenforced>
	
	| agent stateIdx |
	agent := self new.
	agent instVarNamed: #state put: initialState.
	
	^ agent.! !

!OcjAgent methodsFor: 'agent' stamp: 'StefanMarr 7/18/2012 21:56'!
await
	"Block until all actions have been processed (weaker semantics than Clojure's)"
	<omniUnenforced>
	[mailbox isEmpty] 
		whileFalse: [
			(Delay forMilliseconds: 50) wait.
		]! !

!OcjAgent methodsFor: 'initialization' stamp: 'StefanMarr 7/18/2012 21:57'!
initialize
	<omniUnenforced>
	| domain |
	super initialize.
	
	mailbox := SharedQueue new.
	terminate := false.
	
	domain := OcjAgentDomain newFor: self.
	OstMirror setDomainOf: self to: domain.
	
	process := domain newProcessFor: [
		[terminate and: [mailbox isEmpty]] whileFalse: [
			self processIncomingMessages.
		]].
	process name: 'OcjAgent'.
	process resume.! !

!OcjAgent methodsFor: 'agent' stamp: 'StefanMarr 1/7/2012 12:38'!
processIncomingMessages
	| updateBlock |
	updateBlock := mailbox peekWaiting. "Wait until we have it."

	state := updateBlock value: state.
	
	mailbox next. "Removing it only now avoids races with await"! !

!OcjAgent methodsFor: 'agent' stamp: 'StefanMarr 7/18/2012 22:03'!
read
	<omniUnenforced>
	^ state! !

!OcjAgent methodsFor: 'agent' stamp: 'StefanMarr 7/18/2012 22:08'!
send: anUpdateBlock
	<omniUnenforced>
	mailbox nextPut: anUpdateBlock.! !

!OcjAgent methodsFor: 'agent' stamp: 'StefanMarr 7/18/2012 22:08'!
send: anUpdateBlock with: args
	<omniUnenforced>
	self send: [:old | anUpdateBlock value: old value: args ]! !

!OcjAgent methodsFor: 'agent' stamp: 'StefanMarr 7/18/2012 22:08'!
shutdown
	"Make sure that the process loop gets around to see that terminate is set to true.
	 So, we just set it ourselves.
	 No race conditions here, purely sequential."
	<omniUnenforced>
	mailbox nextPut: [:ignore | terminate := true].! !

OcjAgent subclass: #OcjAgentWithImmutability
	instanceVariableNames: 'immutabilityDomain'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'O-Languages-Clojure'!

!OcjAgentWithImmutability methodsFor: 'initialization' stamp: 'StefanMarr 7/18/2012 21:58'!
initialize
	<omniUnenforced>
	super initialize.
	immutabilityDomain := OcjImmutableDomain new.! !

!OcjAgentWithImmutability methodsFor: 'agent' stamp: 'StefanMarr 1/6/2012 15:00'!
processIncomingMessages
	super processIncomingMessages.
	immutabilityDomain adoptObject: state.! !

Object subclass: #CSPiChannel
	instanceVariableNames: 'cell inDomain outDomain'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'O-Languages-CSPi'!

!CSPiChannel class methodsFor: 'instance creation' stamp: 'StefanMarr 8/2/2012 12:52'!
in: domainUsingIn out: domainUsingOut
	| channel |
	channel := self new.
	channel instVarNamed: #inDomain  put: domainUsingIn.
	channel instVarNamed: #outDomain put: domainUsingOut.
	^ channel! !

!CSPiChannel methodsFor: 'channel ends' stamp: 'StefanMarr 8/2/2012 12:52'!
in
	^ CSPiChannelIn newWith: cell for: inDomain.! !

!CSPiChannel methodsFor: 'initialization' stamp: 'StefanMarr 1/9/2012 22:16'!
initialize
	cell := CSPiChannelCell new.! !

!CSPiChannel methodsFor: 'channel ends' stamp: 'StefanMarr 8/2/2012 12:52'!
out
	^ CSPiChannelOut newWith: cell for: outDomain.! !

Object subclass: #CSPiChannelCell
	instanceVariableNames: 'buffer sendSema receiveSema'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'O-Languages-CSPi-Reflection'!

!CSPiChannelCell methodsFor: 'initialization' stamp: 'StefanMarr 1/9/2012 22:11'!
initialize
	sendSema := Semaphore new.
	receiveSema := Semaphore new.! !

!CSPiChannelCell methodsFor: 'cell buffer' stamp: 'StefanMarr 1/9/2012 22:14'!
receive
	| result |
	sendSema wait.
	result := buffer.
	receiveSema signal.
	^ result! !

!CSPiChannelCell methodsFor: 'cell buffer' stamp: 'StefanMarr 1/9/2012 22:12'!
send: aValue
	buffer := aValue.
	sendSema signal.
	receiveSema wait.! !

Object subclass: #CSPiChannelIn
	instanceVariableNames: 'cell domain'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'O-Languages-CSPi-Reflection'!

!CSPiChannelIn class methodsFor: 'instance creation' stamp: 'StefanMarr 8/2/2012 12:55'!
newWith: aCell for: domain
	| end |
	end := CSPiChannelIn new.
	end instVarNamed: #cell   put: aCell.
	end instVarNamed: #domain put: domain. 
	^ end.! !

!CSPiChannelIn methodsFor: 'message handling' stamp: 'StefanMarr 8/2/2012 13:07'!
send: aValue
	<omniUnenforced>
	cell send: aValue! !

Object subclass: #CSPiChannelOut
	instanceVariableNames: 'cell domain'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'O-Languages-CSPi-Reflection'!

!CSPiChannelOut class methodsFor: 'instance creation' stamp: 'StefanMarr 8/2/2012 12:55'!
newWith: aCell for: domain
	| end |
	end := CSPiChannelOut new.
	end instVarNamed: #cell put: aCell.
	end instVarNamed: #domain put: domain.
	^ end! !

!CSPiChannelOut methodsFor: 'message handling' stamp: 'StefanMarr 8/2/2012 13:07'!
receive
	<omniUnenforced>
	| result |
	result := cell receive.
	domain adoptObject: result.
	^ result! !

Object subclass: #OcjAgentExample
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'O-Languages-Clojure'!

!OcjAgentExample class methodsFor: 'example' stamp: 'StefanMarr 1/7/2012 12:40'!
example
	| agent |
	agent := OcjAgent new: 1.

	[ 1 to: 10 do: [:i|
		agent send: [:old |
			Transcript log: '[agent-1] old: ', old asString, ' iteration: ', i asString.
			old + 1].
		(Delay forSeconds: 1) wait.
	]] forkNamed: 'agent-1'.

	[ 1 to: 10 do: [:i|
		agent send: [:old |
			Transcript log: '[agent-2] old: ', old asString, ' iteration: ', i asString.
			old + 1].
		(Delay forMilliseconds: 800) wait.
	]] forkNamed: 'agent-2'.

	[ 1 to: 20 do: [:i|
		agent send: [:old |
			Transcript log: '[agent-3] old: ', old asString, ' iteration: ', i asString.
			old + 1].
		(Delay forMilliseconds: 512) wait.
	]] forkNamed: 'agent-3'.

	Transcript log: 'Started all agent tester processes.'.! !
LRSTMDomain initialize!
